"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
// The MIT License (MIT)
//
// Copyright (c) 2014 Yannick Croissant
// Copyright (c) 2021 Steve Dignam
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
const experimental_utils_1 = require("@typescript-eslint/experimental-utils");
const pragmaUtil = __importStar(require("../pragma"));
const jsx_1 = require("../jsx");
const util = __importStar(require("../utils"));
function first(array) {
    return array[0];
}
function isJSXText(node) {
    return !!node && node.type === "JSXText";
}
function isOnlyWhitespace(text) {
    return text.trim().length === 0;
}
function isNonspaceJSXTextOrJSXCurly(node) {
    return ((isJSXText(node) && !isOnlyWhitespace(node.raw)) ||
        node.type === "JSXExpressionContainer");
}
/// Support <Foo bar={<>foo bar buzz</>} />
function isFragmentWithOnlyTextAndIsNotChild(node) {
    var _a, _b;
    return (node.children.length === 1 &&
        isJSXText(node.children[0]) &&
        !(((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === "JSXElement" || ((_b = node.parent) === null || _b === void 0 ? void 0 : _b.type) === "JSXFragment"));
}
function trimLikeReact(text) {
    const leadingSpaces = /^\s*/.exec(text);
    if (leadingSpaces == null) {
        return text;
    }
    const trailingSpaces = /\s*$/.exec(text);
    if (trailingSpaces == null) {
        return text;
    }
    const start = leadingSpaces[0].includes("\n") ? leadingSpaces[0].length : 0;
    const end = trailingSpaces[0].includes("\n")
        ? text.length - trailingSpaces[0].length
        : text.length;
    return text.slice(start, end);
}
/**
 * Test if node is like `<Fragment key={_}>_</Fragment>`
 */
function isKeyedElement(node) {
    return (node.type === "JSXElement" &&
        node.openingElement.attributes &&
        node.openingElement.attributes.some(jsx_1.isJSXAttributeKey));
}
function containsCallExpression(node) {
    return (node != null &&
        node.type === "JSXExpressionContainer" &&
        node.expression &&
        node.expression.type === "CallExpression");
}
function isJSXFragmentWithEmptyExpr(node) {
    return (node.type === "JSXFragment" &&
        node.children.length === 1 &&
        node.children[0].type === "JSXExpressionContainer" &&
        node.children[0].expression.type === experimental_utils_1.AST_NODE_TYPES.JSXEmptyExpression);
}
/**
 * Test whether a node is an padding spaces trimmed by react runtime.
 */
function isPaddingSpaces(node) {
    return (isJSXText(node) && isOnlyWhitespace(node.raw) && node.raw.includes("\n"));
}
/**
 * Test whether a JSXElement has less than two children, excluding paddings spaces.
 */
function hasLessThanTwoChildren(node) {
    if (!node || !node.children) {
        return true;
    }
    const nonPaddingChildren = node.children.filter((child) => !isPaddingSpaces(child));
    if (nonPaddingChildren.length < 2) {
        return !containsCallExpression(first(nonPaddingChildren));
    }
    return false;
}
function isChildOfHtmlElement(node) {
    var _a, _b;
    return (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === "JSXElement" &&
        ((_b = node.parent) === null || _b === void 0 ? void 0 : _b.openingElement.name.type) === "JSXIdentifier" &&
        /^[a-z]+$/.test(node.parent.openingElement.name.name));
}
function isChildOfComponentElement(node, { reactPragma, fragmentPragma }) {
    var _a;
    return (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === "JSXElement" &&
        !isChildOfHtmlElement(node) &&
        !jsx_1.isFragment(node.parent, reactPragma, fragmentPragma));
}
function canFix(node, pragmas) {
    var _a, _b;
    // Not safe to fix fragments without a jsx parent.
    if (isJSXFragmentWithEmptyExpr(node)) {
        return true;
    }
    if (!(((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === "JSXElement" || ((_b = node.parent) === null || _b === void 0 ? void 0 : _b.type) === "JSXFragment")) {
        // const a = <></>
        if (node.children.length === 0) {
            return false;
        }
        // const a = <>cat {meow}</>
        if (node.children.some(isNonspaceJSXTextOrJSXCurly)) {
            return false;
        }
    }
    // Not safe to fix `<Eeee><>foo</></Eeee>` because `Eeee` might require its children be a ReactElement.
    if (isChildOfComponentElement(node, pragmas)) {
        return false;
    }
    return true;
}
function getFix(node, context, pragmas) {
    if (!canFix(node, pragmas)) {
        return undefined;
    }
    if (isJSXFragmentWithEmptyExpr(node)) {
        return function fix(fixer) {
            return fixer.replaceText(node, "<></>");
        };
    }
    return function fix(fixer) {
        const opener = node.type === experimental_utils_1.AST_NODE_TYPES.JSXElement
            ? node.openingElement
            : node.openingFragment;
        const closer = node.type === experimental_utils_1.AST_NODE_TYPES.JSXElement
            ? node.closingElement
            : node.closingFragment;
        const childrenText = opener.type === experimental_utils_1.AST_NODE_TYPES.JSXOpeningElement && opener.selfClosing
            ? ""
            : context
                .getSourceCode()
                .getText()
                .slice(opener.range[1], closer === null || closer === void 0 ? void 0 : closer.range[0]);
        return fixer.replaceText(node, trimLikeReact(childrenText));
    };
}
function hasSafeInnerExpr(node) {
    // allow <></>
    if (node.children.length === 0) {
        return false;
    }
    const nonPaddingChildren = node.children.filter((child) => !isPaddingSpaces(child));
    // allow <> {foo} </>
    if (nonPaddingChildren.length === 1 &&
        nonPaddingChildren[0].type === experimental_utils_1.AST_NODE_TYPES.JSXExpressionContainer &&
        !isChildOfHtmlElement(node)) {
        return false;
    }
    return true;
}
function checkNode(node, context, pragmas) {
    if (isKeyedElement(node)) {
        return;
    }
    if (isJSXFragmentWithEmptyExpr(node) ||
        (hasLessThanTwoChildren(node) &&
            !isFragmentWithOnlyTextAndIsNotChild(node) &&
            hasSafeInnerExpr(node))) {
        context.report({
            node,
            messageId: "NeedsMoreChidren",
            fix: getFix(node, context, pragmas),
        });
    }
    if (isChildOfHtmlElement(node)) {
        context.report({
            node,
            messageId: "ChildOfHtmlElement",
            fix: getFix(node, context, pragmas),
        });
    }
}
exports.default = util.createRule({
    name: "jsx-no-useless-fragment",
    meta: {
        type: "suggestion",
        fixable: "code",
        docs: {
            description: "Disallow unnecessary fragments",
            category: "Best Practices",
            recommended: "error",
        },
        messages: {
            NeedsMoreChidren: "Fragments should contain either an expression or at least two children.",
            ChildOfHtmlElement: "Passing a fragment to an HTML element is useless.",
        },
        schema: [],
    },
    defaultOptions: [],
    create(context) {
        const reactPragma = pragmaUtil.getFromContext(context);
        const fragmentPragma = pragmaUtil.getFragmentFromContext(context);
        return {
            JSXElement(node) {
                if (jsx_1.isFragment(node, reactPragma, fragmentPragma)) {
                    checkNode(node, context, { reactPragma, fragmentPragma });
                }
            },
            JSXFragment(node) {
                checkNode(node, context, { reactPragma, fragmentPragma });
            },
        };
    },
});
//# sourceMappingURL=jsx-no-useless-fragment.js.map